use strict;
use IO::Scalar;
use HTML::Template;
use Carp;
use PureMessage::Manager::Config qw(%config);
use PureMessage::Version qw($PREFIX);
use Net::Domain qw(hostfqdn);
use PureMessage::Test::Pidfile;

do '../web-lib.pl';
&init_config();
load_theme_library();

ReadParse();

# forward decl
sub test_error($);

our %in;
#our %text;
our $log_dir = $ENV{HOME} || "./";
my $log_file = $config{log_to};
my $pmx = "$PREFIX/etc/init.d/pmx-milter"; 
my $SAMPLES_PATH = "$PREFIX/etc/data/samples";
my $pmx_test_log_file = "$log_dir/pmx_test_log.txt";
my $pmx_load_file = "$log_dir/pmx_load.txt";
my $pmx_thput_file = "$log_dir/pmx_thput.txt";
my $pmx_progress_file = "$PREFIX/lib/webmin/pmx-test/progress.html";
my $LOCAL_PMX = $PREFIX; # FIXME, change the script's behavior if pmx is not on the local box
my $peer = "123.123.123.123";
my $use_milter = ($in{c_test_sendmail} and $in{c_test_sendmail}=~/^on/) || $in{r_test_sendmail};

# globals that are defined by each individual test
our $test_tmpl;
our $params;
our @errors;

{
    my ($name, $value);
    sub v_error {
        my $reason = shift;
        
        my $prefix = "The value entered for '".html_escape($name)."' ('".html_escape($value)."'), ";
        $reason = $prefix.$reason;
        raise_error("error_input_validation", $reason);
        0;
    }
    
    sub validate_integer {
        $value = shift;
        $name = shift;
        my %params = @_;
        
        $value =~ s/\s*(.*?)\s*/$1/;
        $value =~ /^[-\d]+$/ or return v_error("is not an integer");
        
        if (exists $params{min}) {
            ($value >= $params{min}) or return v_error("should be larger than $params{min}");
        }
        if (exists $params{max}) {
            ($value <= $params{max}) or return v_error("should be smaller than $params{max}");
        }
        1;
    }

}

# text inputs
sub reload_text_inputs {
    my $params = shift;
    for my $element (@_) {
        $params->{$element} = html_escape($in{$element}) if exists $in{$element};
    }
}

# select inputs
sub reload_select_inputs {
  my $params = shift;  
  for my $element (@_) {
        for my $node (@{$params->{$element}}) {
            $node->{SEL} = '' if exists $in{$element};
            $node->{SEL} = 'selected' if exists $in{$element} and $in{$element} eq  $node->{VALUE};
        }
    }  
}

sub get_sender_addr {
    my $user = $ENV{USER} || $ENV{LOGNAME} || 'pmx';
    my $host = hostfqdn() || 'localhost';
    return $user.'@'.$host;
}

# add a <br> wherever there is a newline
sub n2br {
    (my $str = shift) =~ s/\n/<br>\n/gm;
    return $str;
}

sub get_tmpl_milters {
    my $selected = 1;
    my @milters;
    for (sort keys %{$config{milter}}) {
        push @milters, {VALUE=>$_,LABEL=>$_,SEL=>$selected};
        $selected = 0;
    }
    return \@milters;
}

sub set_tmpl {
    my $filename = shift;
    $test_tmpl = HTML::Template->new(filename => $filename);
}

sub create_header {
    my $output;
    
    if ($params->{SHOW_REFRESH}) {
        $params->{HEADER} = theme_header_str($params->{TITLE}, undef, 0, 1,'','','',
            "<meta HTTP-EQUIV=\"refresh\" CONTENT=\"$params->{REFRESH}; URL=$params->{URL}\">");
    } 
    else {
        $params->{HEADER} = theme_header_str($params->{TITLE}, undef, 0, 1);
    }
    
}

sub update {
    my $progfile = $params->{PROGFILE};
    my $tmpfile = "$progfile-$$.tmp";
    $test_tmpl->param($params);
    open (TMPFILE, ">$tmpfile") or die "$tmpfile: $!";
    print TMPFILE $test_tmpl->output;
    close TMPFILE;
    rename $tmpfile, $progfile or die "$tmpfile: $!";
}

sub passed_dlg_str {
    return <<EOT;
<div class=passed>
<table width="100%">
<tr>
<td width="26" valign="top"><img src="/images/check.gif"></td>
<td>All tests passed.<br></td>
</tr>
</table>
</div>
EOT
}

sub stat_dlg_str {
    my $stat_str = shift;
    return <<EOT;
<div class=passed>
<table width="100%">
<tr>
<td width="26" valign="top"><img src="/images/check.gif"></td>
<td>Test Complete.  $stat_str</td>
</tr>
</table>
</div>
EOT
}

my %status_icons = (
testing => '<img src="/images/current.gif">',
passed => '<img src="'.get_default_img('check').'">',
failed => '<img src="'.get_default_img('exclamation').'">',
skipped => '<img src="'.get_default_img('dash').'">',
);

# fill in all fields with this value
sub fill_in_checklist {
    my $status = shift;
    for my $node (@{$params->{CHECKLIST}}) {
        if (!$node->{status}) {
            $node->{status} = $status;
            $node->{icon} = $status_icons{$status};
        }
    }
}

sub cleanup {
    $params->{SHOW_REFRESH} = 0;
    my $err_str = error_dlg_str();

    if ($err_str) {
        $params->{FOOTER} = $err_str;
    }
    elsif (!$params->{FOOTER}) {
        $params->{FOOTER} = passed_dlg_str();
    }
    $params->{FOOTER} .= "<p>&nbsp;</p><a href=".$params->{SCRIPT}."><img border=0 src=/images/back.gif></a>";
    $params->{CANCEL_LINK} = '';
    create_header();
    fill_in_checklist('skipped') if (exists $params->{CHECKLIST});
    update();
    exit;
}

sub set_output_files {
    my $testn = shift;
    my $pid = shift;
    my $time = shift || time;
    set_tmpl("tmpl/${testn}_test.html");
    $params->{PROGFILE} = "${testn}-".$pid.'-'.$time.'.html';
    $params->{OUTFILE} = "${testn}-".$pid.'-'.$time.'.out';
    $params->{PIDFILE} = "${testn}-".$pid.'-'.$time.'.pid';
}

sub parent_finish {
    set_output_files(@_);
    create_header();
    update();
    $params->{SHOW_REFRESH} = 1;
    $params->{REFRESH} = 0;
    $params->{URL} = './'.$params->{PROGFILE};
    create_header();
    $params->{HEADER} = "Content-Type: text/html\n\n".$params->{HEADER};
    $test_tmpl->param($params);
    print $test_tmpl->output;
    exit;
}

sub set_status {
    my $key=shift;
    my $value=shift;
    for my $node (@{$params->{CHECKLIST}}) {
        if ($node->{name} eq $key) {
            $node->{status} = $value;
            $node->{icon} = $status_icons{$value};
            last;
        }
    }
}

sub set_value {
    my $key=shift;
    my $value=shift;
    for my $node (@{$params->{STATUS}}) {
        if ($node->{name} eq $key) {
            $node->{value} = $value;
            last;
        }
    }
}

sub fork_test {
    my $pid;
    
    if (not defined ($pid = fork)){
            die "Could not fork!";
    }
    elsif ($pid) {
            return $pid;
    }
    return $pid;
}

sub child_init {
    # all output to file
    set_output_files(@_);
    $params->{CANCEL_LINK}="<a href='cancel.cgi?from=$params->{SCRIPT}&pidfile=$params->{PIDFILE}'><img border=0 src='/images/cancel.gif'></a>";
    PureMessage::Test::Pidfile->new($params->{PIDFILE})->append();
    my $outfile = $params->{OUTFILE};
    create_header();
    open CHILDOUT, ">$outfile" or die "$outfile: $!";
    open STDERR, ">&CHILDOUT" or die "could not redirect STDERR";
    
    # auto-flush outputs
    select STDERR; $| = 1;
    select CHILDOUT; $| = 1;
    
    # disconnect from webserver
    open (STDIN,"</dev/null");
    close STDOUT;
}

sub test_start {
    print "test_start()\n";
    
    raise_error('error_no_pmx_found', $pmx) unless -e $pmx;
    raise_error('error_pmx_not_exe', $pmx, $ENV{USER}) if !@errors and ! -x $pmx;
    
    return 0 if @errors;
    
    my $out = `$pmx start 2>&1`;
    
    print "test_start(): $out\n";
    
    if ( $out =~ /running/ ) {
        # OK
    }
    elsif ( $out =~ /started\./ ) {
        # OK
    }
    else {
        # failed
        my $err = 1010;
        if ( $out =~ /Compilation failed/ 
          or $out =~ /Can\'t locate/  ) {
            $err = 1008;
        }
        elsif ( $out =~ /No milters in/ ) {
            $err = 1009;
        }
	raise_error($err,$out);
	return 0;
	}
    return 1;
}

sub test_stop {
    print "test_stop()\n";
    my $out = `$pmx stop 2>&1`;
    print "test_stop(): $out\n";
    if ( $out =~ /stopped\./ ) {
        # OK
    }
    else {
        raise_error(1004,$out);
        return 0;
    }
    return 1;
}

sub test_status {
    print "test_status()\n";
    my $out = "";
    foreach (1..10) {
        $out = `$pmx status 2>&1`;
        if ( $out !~ /dead\?/ and $out =~ /pid:/) {
            # OK
            print "test_status(): $out\n";
            return 1;
        }
        sleep 5;
    }
    print "test_status(): $out\n";
    raise_error(1004,$out);
    # failed
    return 0;
}

sub test_milter {
    # Send a test message to milter
    my $milter = $in{sl_milter} || 'Main';
    my $milter_socket = $config{milter}{$milter}{port} || "inet:3366\@localhost";
    
    my $mi_test = PureMessage::Test->new({ use_milter => 1,
					   socket => $milter_socket,
					   to => 'test',
				         });
    
    my $test_id = "MILTERTEST".time();
    my $test_msg = sample_message($test_id,$in{t_from_email});
    
    if (not $mi_test->msg_send(\$test_msg)) {
            raise_error(1005, PureMessage::Test->error());
            return 0;
    }

    my $status = 0;
    foreach (@{$mi_test->{mi_actions}}) {
        $status =1 if $_->[1] =~ /X-PMX-Test-Response/;
    }
    
    if (not $status) {
        raise_error(1005, PureMessage::Test->error());
        return 0;
    }
    return 1;
}

sub test_sendmail {
    # need a new test_id to avoid the confusion
    my $test_id = "SMTPTEST".time();
    my $test_msg = sample_message($test_id);
    my $sm_test = PureMessage::Test->new({ use_milter => 0,
					   host   => $in{t_host},
					   port   => $in{t_port},
					   to     => $in{t_to_user},
					   from   => $in{t_from_email},
				         });
    my $smtp_stat = $sm_test->msg_send(\$test_msg);
									  
    if ($smtp_stat ne "Sent") {
        raise_error(1002, PureMessage::Test->error()."\n$smtp_stat");
        return 0;
    }
    
    unless (find_test_log_entry($test_id)) {
        raise_error(1000);
        return 0;
    }
    
    return 1;
}

sub sample_message {
    my $test_id = shift;
	my $from_user = shift;
    return <<EOT;
X-PMX-Test-Request: $test_id
Subject: [pmx-test] $test_id
From: $from_user


This is a test message sent by PureMessage Manager.

PureMessage is a complete email filtering system for spam 
and virus protection, and corporate policy enforcement. 

Thank you for using PureMessage!

http://www.ActiveState.com/Products/PureMessage

EOT
}

sub find_test_log_entry {
    my $test_id = shift;
    open (my $log, "<".$log_file) || die "Can't open log file";
    my $cp;
    my $i=0;
    for (;;) { # from perlfunc sample
        print ".";
        for ($cp = tell($log); $_ = <$log>; $cp = tell($log)) {
            return 1 if /TEST_ID=$test_id/;
        }
        sleep(1);
        $i++;
        seek($log, $cp, 0);
        return 0 if $i>5;
    }
}

sub send_spam {
    my $location = shift || "$SAMPLES_PATH/spam";
    my $progress_func = shift;
    my $msgs = PureMessage::Test::Messages->new( {'location'=>$location} );
    my $stats = send_test_messages($msgs, $progress_func);
    print("$$ done sending\n");
    return 0 if send_failed($stats);
    raise_error(1006), return 0 if ( $stats->{spam} < 1 );
    printf("Correctly detected %.0f%% of spam.<br>\n", $stats->{spam}/$stats->{total}*100);
    return 1;
}

sub send_virus {
	my $location = shift || "$SAMPLES_PATH/virus"; 
	my $progress_func = shift;
        my $msgs = PureMessage::Test::Messages->new( {'location'=>$location});
        my $stats = send_test_messages($msgs, $progress_func);
	return 0 if send_failed($stats);
	raise_error(1003), return 0 if $stats->{virus} != $stats->{total};
	return 1;
}

sub send_custom {
    my $location = shift || "$SAMPLES_PATH/spam";
    my $progress_func = shift;
    my $msgs;
    eval {
        $msgs = PureMessage::Test::Messages->new( {'location'=>$location} ); # $in{t_custom_msg_folder}
    };
    if ($@) {
        if ($@ =~ /open/) {
            raise_error('error_custom_cant_open_dir',$location);
            return 0;
        }
        elsif ($@ =~ /No messages found/) {
            raise_error('error_custom_no_msgs_found',$location);
            return 0;
        }
        else {
            raise_error('error_unknown',$@);
            return 0;
        }
    }
    my $stats = send_test_messages($msgs, $progress_func);
    print "<br>\nTest ";
    return 0 if send_failed($stats);
    print "COMPLETE<br>\n";
    return 1;
}

sub send_test_messages {
	my $msgs = shift;
	my $progress_func = shift;
	if ( $msgs ) {
		my $test = create_test("PureMessage::Test::WFilters",$progress_func);
		open ( my $fh, ">$pmx_test_log_file" ) 
			|| print "Can't create temporary file $pmx_test_log_file";
		$test->{msgs} = $msgs;
		$test->{clients_min} = $in{t_forks} || 1;  #FIXME: these should be user defn'd
		$test->{timeout} = $in{t_timeout} || 0;
		$test->{report} = {'file'=>$fh, 'progfile'=>$pmx_progress_file };
		my $stats = $test->run();
		print $fh "<br>\nTest ";
		return 0 if send_failed($stats);
		print $fh "COMPLETE<br>\n";
		return $stats;
	}
	return undef;
}

sub send_failed {
    my $stats = shift;
    if ( $stats ) {
        if ( $stats->{total} > 0 ) {
            return 0;
        }
        else {
            raise_error(1007);
            return 1;
        }
    }
    return 1;
}

sub create_test {
        my $milter = $in{sl_milter} || 'Main';
	my $test_type = shift || "PureMessage::Test";
	my $progress_func = shift;
	my $test;
	
	print STDERR "create_test()\n";
	
    if ($use_milter){
         $test = $test_type->new( { use_milter  => 0,
                                    local_pmx=> $LOCAL_PMX,
                                    host        => $in{t_host},
                                    port        => $in{t_port},
                                    to          => $in{t_to_user},
                                    from        => $in{t_from_email},
                                    peer        => $peer,
				    progress_func=> $progress_func,
				    pidfile => $params->{PIDFILE},
				    } );
    }
    else {  
         print "Using direct connection to PureMessage.<br>\n";
         $test = $test_type->new( { use_milter  => 1,
                                    local_pmx=> $LOCAL_PMX,
                                    socket      => $config{milter}{$milter}{port} || "inet:3366\@localhost", #FIXME: this should be capable of working remotely?
                                    to          => 'test',
                                    peer        => $peer,
				    progress_func=>$progress_func,
				    pidfile => $params->{PIDFILE},
				} );
    }
    return $test;
}

sub test_maxload {
    my $progress_func = shift;
    my $max_forks = shift;
    my $int = shift;
	my $location = shift || "$SAMPLES_PATH/spam";
    my $est_test_time = $max_forks * $int + 10;
    my $msgs = PureMessage::Test::Messages->new( 
                {'location'=>$location, 'count'=>10000} );			
				
    my %report;
    $report{granularity} = $int; # this is a granularity for the report in seconds

    open ( my $fh, ">$pmx_load_file" ) 
           || print "Can't create temporary file $pmx_load_file<br>\n";
    $report{file} = $fh;

    my $test = create_test("PureMessage::Test::WLoad",$progress_func);
    $test->{msgs} = $msgs;
    $test->{clients_min} = 1;
    $test->{clients_max} = $max_forks;
    $test->{interval} = $int;
    $test->{report} = \%report;
    $test->{timeout} = $est_test_time;
    print STDERR "about to run()\n";
    my $stats = $test->run; 
	if (!$stats) {
		print STDERR "test->run() returned undef\n";
	}
	else {
		print STDERR "test->run() returned ok\n";
	}

    close $report{file};
    return if send_failed($stats);
    return "<br>\n<br>\nAverage throughput: " 
      . sprintf("%.1f", $stats->{total} / ( $stats->{time}-$stats->{start} ) * 60 ) 
      . " messages per minute\n";
}
	  
sub test_performance {
    my $progress_func = shift;
    my $forks = shift;
    my $test_time = shift;
    my $location = shift || "$SAMPLES_PATH/spam";
    my $msgs = PureMessage::Test::Messages->new( 
                {'location'=>$location, 'count'=>10000} );
    my %report;
    $report{granularity} = int($test_time/50); # this is a granularity for the report in seconds
    $report{granularity} = 5 if $report{granularity} < 5;

    open ( my $fh, ">$pmx_thput_file" ) 
           || print "Can't create temporary file $pmx_thput_file";
    $report{file} = $fh;

    my $test = create_test("PureMessage::Test::WPerformance",$progress_func);
    $test->{msgs} = $msgs;
    $test->{clients_min} = $forks;
    $test->{report} = \%report;
    $test->{timeout} = $test_time;
    my $stats = $test->run; 

    close $report{file};
    return if send_failed($stats);
    return "<br>\n<br>\nAverage throughput: " 
      . sprintf("%.1f", $stats->{total} / ( $stats->{time}-$stats->{start} ) * 60 ) 
      . " messages per minute\n";
}

sub test_error($) {
	my $err = shift;
	my %troubleshooter = (

1000 => <<EOT,

Sendmail cannot connect to PureMessage on this server.

Please verify that:
1. Sendmail\'s Milter support is enabled.
2. The filter is correctly configured in the sendmail.cf file,
   with correct PureMessage port and hostname values.
See "Configuring Sendmail" in the Installation section of the PureMessage 
User Guide (or in pmx-setup) for instructions.
  
EOT

1001 => <<EOT,

Cannot determine PureMessage port. Please verify that PureMessage is
installed on this server.

EOT

1002 => <<EOT,

Can\'t send the message to Sendmail. 

Please verify that:
1. Sendmail is running.
2. The specified recipient exists.
3. The correct smtp hostname and port have been specified.
4. PureMessage is running.

EOT

1003 => <<EOT,
Virus detection problem.

Please verify that:
1. The VirusCheck filter is installed. 
2. The VirusCheck filter is enabled in etc/pmx.conf.
3. The Virus scanner license is valid (use "pmx version" to 
   display version and license information).

EOT

1004 => <<EOT,

EOT

1005 => <<EOT,
Connection to PureMessage failed.

Please verify that:
1. PureMessage is running.
2. The specified PureMessage socket address is correct.

EOT

1006 => <<EOT,
Spam detection problem.

Please verify that:
1. The SpamCheck filter is installed. 
2. The SpamCheck filter is enabled in etc/pmx.conf.

EOT

1007 => <<EOT,

Message delivery problems.
There were no messages sent through.
Please correct the errors and try again.

EOT

1008 => <<EOT,
Cannot start PureMessage.

Please verify that:
1. There are no errors in pmx.conf and other config files.
2. If you have <milter Main> section in your pmx.conf file it 
   should contain at least one filter definition.

EOT

1009 => <<EOT,
Cannot start PureMessage.

Please verify that:
1. There is at least one milter section in your pmx.conf file, 
   for example:
   <milter Main>
       port = inet:3366\@localhost
       filter = PureMessage::SpamCheck
   </milter>

EOT

1010 => <<EOT,
Cannot start PureMessage.

Please verify that:
1. The port PureMessage is trying to bind to is not occupied by other programs.

EOT
);
	return exists $troubleshooter{$err} ? n2br($troubleshooter{$err}) : "Unknown error code $err.\n<br>";
}

1;

