package PureMessage::Test;

use strict;

our $VERSION = "5.509";

use POSIX ":sys_wait_h";
use Term::ANSIColor qw(:constants);
use PureMessage::Milter::Run qw(running);
use PureMessage::MessageLog qw(message_log);
use PureMessage::Test::Pidfile;

$| = 1;

my $ERROR="";

sub new {
    my $class = shift;
    my $params = shift;
	
	
    unless ($params->{from}) {
        $params->{from} = $ENV{USER} || getpwuid($<) || $ENV{LOGNAME} ||
	    ( $ERROR = "Can't determine user name" && return undef );
        require Sys::Hostname;
        $params->{from} .= '@' . Sys::Hostname::hostname();
    }
    unless ($params->{to}) {
        $ERROR = "Recipient address is not specified.";
        return undef;
    }
    $params->{host} ||= "localhost";
    $params->{use_milter} ||= 0; # use smtp by default
    $params->{port} ||= $params->{use_milter} ? 3366 : 2525;
    $params->{max_mi_arg} ||= 70;
    $params->{verbose} ||= 0;
    $params->{queue_id} = "g" . $$ . "aaa";
    $params->{queue_id} =~ tr/0-9/A-J/;  # make sure magical increment works
    $params->{queue_id} .= "0";
    $params->{paranoid} ||= 1;
    $params->{test_id} ||= join("-", time, $$);
    $params->{pidfile_o} = PureMessage::Test::Pidfile->new($params->{pidfile}) if $params->{pidfile};
    my $self = bless { %{$params} }, $class;
    $self;
}

sub error { return $ERROR };

# overload to suppress output
sub print_out {
    my $self = shift;
    print join "", @_;
}

sub feed {
    my $self = shift;
    my $start = time();

    $self->{msgs}->reset();
    my %status;
    my $i = 0;
    while ( ! $self->{msgs}->eof 
            and ( ! $self->{stop_time} or time() < $self->{stop_time} ) ) {
        my $res = $self->msg_send( $self->{msgs}->fetch );
        $res =~ s/[\r\n]/ /g;
        $self->evt_sent($res); # do some custom defined logging
        $status{$res}++;
        if ( $self->{paranoid} and $i > 5 and ! defined $status{"Sent"} ) {
            # looks like a fatal error, stop the feed and return
            return \%status;
        } 
        $i++;
    }
    return \%status;
}

sub msg_send {
    my $self = shift;
    my $msg = shift;
    return "Can't read the message" if ! $msg;
    return $self->milter_send($msg) if $self->{use_milter};
    return $self->smtp_send($msg);
}


### Communication protocols ######################################

sub smtp_connect {
    my $self = shift;
    my $force = shift;
    if ( $force ) {
        $self->{smtp}->quit if defined $self->{smtp};
        $self->{smtp} = undef;
    }
    if ( not defined $self->{smtp} ) {
        require Net::SMTP;
        $self->{smtp} = Net::SMTP->new($self->{host}, Port => $self->{port}, Debug => $self->{verbose}) ||
            return undef ;
    }
    return $self->{smtp};
}

sub smtp_send {
    my $self = shift;
    my $msg = shift;
    my $smtp = $self->smtp_connect(1);
    my $stat = eval {
        return "Can't connect to $self->{host}:$self->{port}" if ! $smtp;
        $smtp->mail($self->{from}) || return $smtp->message;
        $smtp->to($self->{to}) || return $smtp->message;
        $smtp->data() || return $smtp->message;
        $smtp->datasend("X-PMX-Test-Request: ".$self->{test_id}."\n") if $self->{test_id};
        $smtp->datasend(${$msg}) || return $smtp->message;
        $smtp->dataend() || return $smtp->message;
        return "Sent";
    };
    if ($@) {
        $ERROR = $@;
        return $@;
    }
    return $stat;
}

sub DESTROY {
    my $self = shift;
    $self->{smtp}->quit if $self->{smtp};
#    warn "DEBUG: DESTROY";
}

sub milter_connect {
    my $self = shift;
    my %params;
    $params{Debug} = $self->{verbose};
    if ( $self->{socket} ) {
        $params{Socket} = $self->{socket};
    }
    else {
        my $kind = ($self->{host} =~ /^\//) ? "Path" : "Host";
        $params{$kind} = $self->{host};
        $params{Port} = $self->{port};
    }
    require ActiveState::Milter;
    eval {$self->{milter} = ActiveState::Milter->new(%params)};
    do { $ERROR = "Milter connection failed: $@"; return undef } unless $self->{milter};
    return $self->{milter};
}

sub milter_send {
    my $self = shift;
    my $msg_h = shift;
    my $msg = "";
    $msg .= "X-PMX-Test-Request: " . $self->{test_id} . "\n" if $self->{test_id};
    $msg .= ${$msg_h};
    
    $self->{mi_actions} = [];
    
    my $m = $self->milter_connect;
    return error() unless $m;
    
    $m->action_cb( sub{ $self->milter_action(@_) } );

    my @connect_args = ("unknownhost");
    if ($self->{peer}) {
	require Socket;
	my $host = $self->{peer};
	my $port = 0;
	$port = $1 if $host =~ s/:(\d+)$//;
	@connect_args = $host;
	if (my $n = Socket::inet_aton($host)) {
	    push(@connect_args, Socket::pack_sockaddr_in($port, $n));
	}
    }

    eval {
	$m->mi_def_macro("connect", "j" => "foo.bar.com", "_" => "[ident]");
	mi_c($m, "connect", @connect_args);

	$m->mi_def_macro("mail_from", "i", $self->{queue_id}++, "{mail_mailer}", "local");
	mi_c($m, "from", "<$self->{from}>");
	mi_c($m, "to", "<$self->{to}>");
	# eat the headers one by one
	while ($msg =~ s/^(\S+)\s*: ?(.*?)\r?\n(?=\r?\n|\S)//s) {
	    mi_c($m, "header", $1, $2);
	}

	$msg =~ s/^\r?\n//;
	mi_c($m, "eoh");

	while (length($msg) > 1024) {
	    mi_c($m, "body", substr($msg, 0, 1024, ""));
	}

	mi_c($m, "eom", $msg);
    };
    my $err = $@;
    $m->mi_quit;
    
    if ($err && $err !~ /^Accepted\b/) {
	chomp($err);
	return $err;
    }
    return "Sent";
}

sub mi_c {
    my $m = shift;
    my $meth = "mi_" . shift;
    my($status, $buf) = $m->$meth(@_);
    if ($status ne "c") {
	die "Tempfailed\n" if $status eq "t";
	die "Accepted\n" if $status eq "a";
	die "Discarded\n" if $status eq "d";
	die "Rejected\n" if $status eq "r";
	die "$buf\n" if $status eq "y";

	die "$meth returned '$status'\n";
    }
}

sub milter_action {
    my $self = shift;
    return if $self->{max_mi_arg} == 0;
    my $m = shift;
    my $action = shift;
    require Data::Dump;
    push @{$self->{mi_actions}}, [$action, join(", ", map Data::Dump::dump($self->shorten($_)), @_)];
}

sub shorten {
    my $self = shift;
    my $str = shift;
    my $max = $self->{max_mi_arg};
    if (length($str) > $max) {
	my $n = $max / 2 - 5;
	$n = 1 if $n < 1;
	substr($str, $n, -$n) = " ... ";
    }
    $str;
}

### High-level functions #######################################

sub spawn_feeding {
    my $self = shift;
    my $id = shift;
    $id = defined($id) ? " #$id" : "";

    my $pid = fork;
    if (!defined($pid)) {
        $self->print_out("ERROR: cannot fork: $!\n");
        return -1;
    }
    elsif ($pid) {
	# the parent
        $self->{pidfile_o}->append($pid) if $self->{pidfile_o};
        return $pid;
    }

    # the child
    my $status = $self->feed();
    $self->print_out("Client$id has finished:                                       \n");

    if (%$status) {
	my $max_width = 0;
	for (keys %$status) {
	    $max_width = length if length > $max_width;
	}
	for my $k (sort keys %$status) {
	    $self->print_out(sprintf "    %-*s %2d messages\n", $max_width+1, "$k:", $status->{$k});
	}
	$self->print_out("\n");
    }
    exit;
}

sub ensure_pmx {
    my $self = shift;
    $self->print_out(BOLD, "\nEnsuring that PureMessage is running...\n", CLEAR);
    system ("pmx start"); # ensure that PureMessage is running
    return 1; # FIXME - can we check status of pmx start here instead?
}

sub run {
    my $self = shift;
    my $milter = shift || "";

    return undef if $self->{local_pmx} && not $self->ensure_pmx();

    my $log_file =  $self->{local_pmx} ? message_log(undef, $milter)
	                               : "/tmp/pmx_test_".time().".tmp";

    $self->write_report_header;

    # touch the log file so that we can start reading it even if it
    # did not exists when we start up.
    my $log;
    open ($log, ">>", $log_file) && close($log);
    undef($log);

    open ($log, $log_file) 
             || die "Can't open log file '$log_file': $!";
    seek($log, 0, 2); # seek to the end

    if ($self->{clients_min} > 1) {
	$self->print_out("\n", BOLD, "Starting $self->{clients_min} clients...\n", CLEAR);
    }
    else {
	$self->print_out("\n", BOLD, "Starting the client...\n", CLEAR);
    }


    $self->{clients} = [];
    $self->{total_clients} = 0;
    $self->{stop_time} = $self->{timeout} ? time + $self->{timeout} : 0;
    foreach (1 .. $self->{clients_min}) {
	my $pid = $self->spawn_feeding($self->{clients_min} > 1 ? $_ : undef);
	if ($pid > 0) {
	    $self->{total_clients}++;
	    push(@{$self->{clients}}, $pid);
	}
    }

    my %stats = (
	start => time,
        total => 0,    # total number of messages

        ok    => 0,    # number of accepted messages
	spam  => 0,    # number of with spam probablility > 50%
	virus => 0,    # number of virus infected messages

        prev_total => 0,
        prev_time  => time,
    );

    my $waiting_printed;
    while ( @{$self->{clients}} ) {
	# pick up children that has terminated
	while ((my $pid = waitpid(-1, WNOHANG)) > 0) {
	    @{$self->{clients}} = grep $_ != $pid, @{$self->{clients}};
	}

	sleep(1);

	# read the log until EOF
	while (<$log>) {
	    # count messages processed by pmx in message_log
	    my($ts, %hash) = PureMessage::MessageLog::split_line($_);
	    if (($hash{'Test-ID'} || "") eq $self->{test_id}) {
		$stats{total}++;
		$stats{virus}++ if exists $hash{v};
		$stats{spam}++ if ($hash{p} || 0) >= 0.5;
		$stats{ok}++ if $hash{a} =~ m,^(a|c)/,;
	    }
	}

	# reset EOF state
	seek($log, tell($log), 0);
	
	my $now = $stats{time} = time;
	$self->evt_idle(\%stats);

	next unless $self->{stop_time} && @{$self->{clients}};

	my $secs_since_stop = $now - $self->{stop_time};
	if ($secs_since_stop > 5 && $waiting_printed) {
	    # Clients should normally stop by themselves, but are now
	    # 5 seconds late.
	    $self->print_out(BOLD, "Test timeout.  Waiting for child processes to exit...",  "              \n", CLEAR);
	    $waiting_printed++;
	}

	if ($secs_since_stop > 15) {
	    # Give up waiting.  Force termination of remaining clients.
	    while ( @{$self->{clients}} ) {
		my $pid = shift @{$self->{clients}};
		next if waitpid($pid, WNOHANG) == -1;

		if (kill(9, $pid)) {
		    $self->print_out("Processes pid=$pid terminated due to the test timeout.\n");
		    waitpid($pid, 0);
		}
		else {
		    print "Can't terminate process pid=$pid.           \n"
			if waitpid($pid, WNOHANG) != -1;
		}
	    }
	    last;
	}
    }

    return \%stats;
}

sub show_progress {
    my $self = shift; 
    my $stats = shift; 
    $self->print_out("Progress: $stats->{total} messages processed in " . ( $stats->{time}-$stats->{start} ) . " seconds\r");
}

sub write_report_header {
#
}

sub evt_idle {
# do nothing in the base class
}

sub evt_sent {
# log results from "feeder" here if PureMessage is not on the local box
}

sub get_pmx_status {
    my $self = shift;
    my $running = running(ps_info => 1, ping => 0);
    my $pmx_stat = "";
    my ($vsize, $threads, $procs) = (0,0,0);
    for my $r (sort keys %$running) {
	my $info = $running->{$r};
        $vsize += $info->{vsize} if $info->{vsize};
        $threads += $info->{threads} if $info->{threads};
        $procs += $info->{procs} if $info->{procs};
    }
    # vmsize threads procs
    $pmx_stat .= $vsize
                 ? sprintf("%.1fMB ", $vsize / 1024 ) 
                 : "  ?";
    $pmx_stat .= $threads
                 ? sprintf("% 3d ", $threads ) 
                 : "  ?";
    $pmx_stat .= $procs
                 ? sprintf("% 3d ", $procs ) 
                 : "  ?";
    return $pmx_stat;
}

sub get_loadavg {
    # linux only for now
    open (my $F, "uptime|") || return "";
    my $load_str = <$F>;
    return sprintf("% 5.2f", $1) if $load_str =~ /load average: (\d+\.\d+),/;
    return "";
}

1;

package PureMessage::Test::Performance;
use base 'PureMessage::Test';

sub write_stats_line {
    my ( $self, $stats ) = @_;
    # write detailed stats to the files
    my $stat_str = sprintf("% 6d  % 6d  % 6d % 7d ", 
                    ( 
                      $stats->{time}-$stats->{start} ), 
                      $stats->{total}, 
                      $stats->{total} - $stats->{prev_total},
                      ($stats->{total} - $stats->{prev_total}) / ($self->{report}->{granularity}) * 60,
                       );
    my $fh = $self->{report}->{file};
    print $fh $stat_str . " ";
    print $fh $self->get_loadavg() . " ";
    print $fh $self->get_pmx_status();
    print $fh "\n";

}

sub evt_idle {
    my $self = shift;
    my $stats = shift;
    if ($self->{report}->{granularity} and 
           $stats->{time}-$stats->{prev_time} >= $self->{report}->{granularity} ) {
        $self->write_stats_line($stats);
        $stats->{prev_total} = $stats->{total};
        $stats->{prev_time} = $stats->{time};
    }
    $self->show_progress($stats);
}

sub write_report_header {
    my $self = shift;
    my $fh = $self->{report}->{file};
    return unless $fh;
    print $fh "PureMessage Throughput Test Report\n\nDetailed stats for each $self->{report}->{granularity} seconds of processing time.\n\n";
    print $fh <<EOT;

Time    Total   New    Average  Load  PureMessage:
(sec.)  msgs.   msgs.  per min. avg.  vsize  thrds procs
------- ------- ------ -------- ----- -------------------
EOT
}


1;

package PureMessage::Test::WPerformance;
use base 'PureMessage::Test';

sub write_stats_line {
    my ( $self, $stats ) = @_;
    # write detailed stats to the files
    my $stat_str = sprintf("% 6d  % 6d  % 6d % 7d ", 
                    ( 
                      $stats->{time}-$stats->{start} ), 
                      $stats->{total}, 
                      $stats->{total} - $stats->{prev_total},
                      ($stats->{total} - $stats->{prev_total}) / ($self->{report}->{granularity}) * 60,
                       );
    my $fh = $self->{report}->{file};
    print $fh $stat_str . " ";
    print $fh $self->get_loadavg() . " ";
    print $fh $self->get_pmx_status();
    print $fh "\n";

}

sub evt_idle {
    my $self = shift;
    my $stats = shift;
    if ($self->{report}->{granularity} and 
           $stats->{time}-$stats->{prev_time} >= $self->{report}->{granularity} ) {
        $self->write_stats_line($stats);
        $stats->{prev_total} = $stats->{total};
        $stats->{prev_time} = $stats->{time};
    }
    $stats->{nclients} = $self->{total_clients};
    my $progress_func = $self->{progress_func};
    $self->{progress_func}->($stats);
}

sub write_report_header {
    my $self = shift;
    my $fh = $self->{report}->{file};
    return unless $fh;
    print $fh "PureMessage Throughput Test Report\n\nDetailed stats for each $self->{report}->{granularity} seconds of processing time.\n\n";
    print $fh <<EOT;

Time    Total   New    Average  Load  PureMessage:
(sec.)  msgs.   msgs.  per min. avg.  vsize  thrds procs
------- ------- ------ -------- ----- -------------------
EOT
}


1;

package PureMessage::Test::Filters;
use base 'PureMessage::Test';
use Term::ANSIColor qw(:constants);

sub show_progress { 
    my $self = shift;
    my $stats = shift; 
    print UNDERLINE, "Total";
    print CLEAR, ": $stats->{total}, "; 
    print UNDERLINE, "Accepted";
    print CLEAR, ": $stats->{ok}, ";
    print UNDERLINE, "Spam (p>50%)";
    print CLEAR, ": $stats->{spam}, ";
    print UNDERLINE, "Virus";
    print CLEAR, ": $stats->{virus}, ";
    print UNDERLINE, "Time";
    print CLEAR, ": ".( $stats->{time}-$stats->{start} ) . "s\r";
};

sub evt_idle {
    my $self = shift;
    my $stats = shift;
    $self->show_progress($stats);
}

1;

package PureMessage::Test::WFilters;
use base 'PureMessage::Test';

sub evt_idle {
    my $self = shift;
    my $stats = shift;
	my $progress_func = $self->{progress_func};
    $self->{progress_func}->($stats);
}

1;

package PureMessage::Test::Load;
use base 'PureMessage::Test';
use Term::ANSIColor qw(:constants);

sub write_stats_line {
    my ( $self, $stats ) = @_;
    # write detailed stats to the files
    my $stat_str = sprintf("% 6d  % 6d  % 6d  % 6d % 7d ", 
                    ( $self->{total_clients},
                      $stats->{time}-$stats->{start} ), 
                      $stats->{total}, 
                      $stats->{total} - $stats->{prev_total},
                      ($stats->{total} - $stats->{prev_total}) / ($self->{report}->{granularity}) * 60,
                       );
    my $fh = $self->{report}->{file};
    print $fh $stat_str . " ";
    print $fh $self->get_loadavg() . " ";
    print $fh $self->get_pmx_status();
    print $fh "\n";

}

sub evt_idle {
    my $self = shift;
    my $stats = shift;
    if ($self->{report}->{granularity} and 
           $stats->{time}-$stats->{prev_time} >= $self->{report}->{granularity} ) {
        $self->write_stats_line($stats);
        $stats->{prev_total} = $stats->{total};
        $stats->{prev_time} = $stats->{time};

        # start another feed
        if ( $self->{total_clients} < $self->{clients_max} ) {
            $self->{total_clients}++;
            print BOLD, "Starting client #$self->{total_clients}...",
                  "                          \n", CLEAR;
            push @{$self->{clients}}, $self->spawn_feeding($self->{total_clients})
        }
    }
    $self->show_progress($stats);
}

sub write_report_header {
    my $self = shift;
    my $fh = $self->{report}->{file};
    return unless $fh;
    print $fh "PureMessage Load Test Report\n\nDetailed stats for each $self->{report}->{granularity} seconds of processing time.\n\n";
    print $fh <<EOT;
 
Clients Time    Total   New    Average  Load  PureMessage:
        (sec.)  msgs.   msgs.  per min. avg.  vsize  thrds procs
------- ------- ------- ------ -------- ----- -------------------
EOT
}

1;

package PureMessage::Test::WLoad;
use base 'PureMessage::Test';

sub write_stats_line {
    my ( $self, $stats ) = @_;
    # write detailed stats to the files
    my $stat_str = sprintf("% 6d  % 6d  % 6d  % 6d % 7d ", 
                    ( $self->{total_clients},
                      $stats->{time}-$stats->{start} ), 
                      $stats->{total}, 
                      $stats->{total} - $stats->{prev_total},
                      ($stats->{total} - $stats->{prev_total}) / ($self->{report}->{granularity}) * 60,
                       );
    my $fh = $self->{report}->{file};
    print $fh $stat_str . " ";
    print $fh $self->get_loadavg() . " ";
    print $fh $self->get_pmx_status();
    print $fh "\n";

}

sub evt_idle {
    my $self = shift;
    my $stats = shift;
    if ($self->{report}->{granularity} and 
           $stats->{time}-$stats->{prev_time} >= $self->{report}->{granularity} ) {
        $self->write_stats_line($stats);
        $stats->{prev_total} = $stats->{total};
        $stats->{prev_time} = $stats->{time};

        # start another feed
        if ( $self->{total_clients} < $self->{clients_max} ) {
            $self->{total_clients}++;
            print "Starting client #$self->{total_clients}...",
                  "                          \n";
            push @{$self->{clients}}, $self->spawn_feeding($self->{total_clients})
        }
    }
    $stats->{nclients} = $self->{total_clients};
    my $progress_func = $self->{progress_func};
    $self->{progress_func}->($stats);
}

sub write_report_header {
    my $self = shift;
    my $fh = $self->{report}->{file};
    return unless $fh;
    print $fh "PureMessage Load Test Report\n\nDetailed stats for each $self->{report}->{granularity} seconds of processing time.\n\n";
    print $fh <<EOT;
 
Clients Time    Total   New    Average  Load  PureMessage:
        (sec.)  msgs.   msgs.  per min. avg.  vsize  thrds procs
------- ------- ------- ------ -------- ----- -------------------
EOT
}

1;
